package client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.concurrent.ConcurrentHashMap;

import net.AppServerSocket;
import net.AppSocket;
import net.ByteOp;
import net.ConnectionInfo;
import net.TimeoutLimitException;
import net.Transport;
import server.LoginException;
import server.RepositorioUsers;
import server.UnnavailableLoginException;
import server.User;
import voip.Voice;

public class Client extends Thread{
	ServerConnection server;
	private boolean logado;
	private boolean conectado;
	private boolean vive;
	private FriendThread ft;
	private AppServerSocket requestSocket;
	private int nextId;
	private ConcurrentHashMap<InetAddress, AppSocket> tabelaDeSockets;
	private ConcurrentHashMap<InetAddress, ChatThread> listaDeThreads;
	private Transport trans;
	private ClientGuiListener listener;
	private String login;
	public Client(ServerConnection server, Transport t) throws IOException,
	ClassNotFoundException{
		conectado = false;
		logado = false;
		vive = true;
		this.server = server;
		ft = new FriendThread(this);
		requestSocket = new AppServerSocket(t,0);
		tabelaDeSockets = new ConcurrentHashMap<InetAddress, AppSocket>();
		listaDeThreads = new ConcurrentHashMap<InetAddress, ChatThread>();
		nextId = 2;
		this.trans = t;
	}
	public ConcurrentHashMap<InetAddress, AppSocket> getTabelaDeSockets(){
		return tabelaDeSockets;
	}
	public ConcurrentHashMap<InetAddress, ChatThread> getListaDeThreads(){
		return listaDeThreads;
	}
	public void addGuiListener(ClientGuiListener listener){
		this.listener = listener;
		ft.addListener(listener);
	}
	public void cadastrar(User usuario) throws IOException, TimeoutLimitException, UnnavailableLoginException, ServerException{
		this.connect();
		server.cadastrar(usuario);
	}
	public void connect() throws TimeoutLimitException{
	
		if(!conectado){

			server.connect();
			
			conectado = true;
		}

	}
	public Hashtable<String, ConnectionInfo> getFriendList() throws TimeoutLimitException, IOException, ClassNotFoundException, ServerException{
		return server.getOnlineFriends();
	}
	public void logar(String login, String senha) throws TimeoutLimitException, LoginException{
		if(!logado){
			if(!conectado){
				this.connect();
			}
			server.logar(login, senha);
			logado = true;
			ft.start();
		}
	}
	public ArrayList<String> getAmigos() throws TimeoutLimitException, IOException, ClassNotFoundException, ServerException{
		return server.getFriends();
	}
	public void adicionarAmigo(String login) throws LoginException, ServerException, TimeoutLimitException, IOException, UnnavailableLoginException{
		server.addFriend(login);
		
	}

	public void logout() throws TimeoutLimitException{
		this.setOffline();
	}
	public void setOffline(){
		System.out.println("set offline");
		this.logado = false;
		this.conectado = false;
		try {
			server.logout();
		} catch (TimeoutLimitException e) {
			
		}
		ft.kill();
		Enumeration<ChatThread> chatThreads = listaDeThreads.elements();
	
		while(chatThreads.hasMoreElements()){
			chatThreads.nextElement().kill();
		}
		
		ft = new FriendThread(this);
		server.newConnection(new AppSocket(trans,1));
	}
	public ClientGuiListener getListener(){
		return this.listener;
	}
	public void run(){
		AppSocket sock = null;
		while(vive){
			ConnectionInfo ci = requestSocket.receiveRequest();
			System.out.println("Recebeu requisio de: "+ci.toString());
			if(!tabelaDeSockets.containsKey(ci.getIp())){
				System.out.println("No estava na tabela");
				sock = new AppSocket(trans, nextId);
				nextId++;
				sock.match(ci);
				tabelaDeSockets.put(ci.getIp(), sock);
				ChatThread thread = new ChatThread(this, sock, ci, this.listener);
				thread.addListener(this.listener);
				listaDeThreads.put(ci.getIp(), thread);
			}else{
				System.out.println("J estava na tabela");
				sock = new AppSocket(trans, nextId);
				nextId++;
				sock.match(ci);
				listaDeThreads.get(ci.getIp()).kill();
				ChatThread thread = new ChatThread(this, sock, ci, this.listener);
				thread.addListener(this.listener);
				listaDeThreads.put(ci.getIp(), thread);
				tabelaDeSockets.put(ci.getIp(), sock);
				
			}
			if(sock != null){
				sock.acceptConnection();
			}
		}
	}
	public void chamadaDeVoz(ConnectionInfo ci, ClientGuiListener listener) throws TimeoutLimitException, SocketException{
		byte[] pacote = {1};
		AppSocket sock = this.connectToClient(ci, listener);
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
		sock.send(pacote);
		new Voice(ci, new DatagramSocket(3000)).comeco();
	}
	public void sendMsg(String msg, ConnectionInfo ci, ClientGuiListener listener) throws TimeoutLimitException{
		byte[] rawMsg = null;
		try {
			rawMsg = msg.getBytes("UTF-16");
		} catch (UnsupportedEncodingException e) {
			rawMsg = msg.getBytes();
			System.out.println("Charset no suportado");
		} 
		AppSocket sock = this.connectToClient(ci, listener);
		ByteBuffer pacote = ByteBuffer.allocate(rawMsg.length+5);
		pacote.put((byte)0);
		pacote.put(ByteOp.intToByte(rawMsg.length));
		pacote.put(rawMsg);
		
		pacote.flip();
		if(sock == null){
			sock = tabelaDeSockets.get(ci.getIp());
		}
		System.out.println("-------------"+rawMsg.length);
		try{
			sock.send(pacote.array());
		}catch(TimeoutLimitException e){
			listaDeThreads.get(ci.getIp()).kill();
			sock = this.connectToClient(ci, listener);
			if(sock != null){
				sock.send(pacote.array());
			}else{
				System.out.println("NAO TA REMOVENDO DIREITO");
			}
		}
	}
	public void convidaParaGroup(ConnectionInfo ci, Hashtable<ConnectionInfo, String> tabelaDeParticipantes, ClientGuiListener listener, ClientGuiListener groupListener) throws IOException, TimeoutLimitException{
//		ByteArrayOutputStream baos = new ByteArrayOutputStream();
//		ObjectOutputStream oos = new ObjectOutputStream(baos);
//		oos.writeObject(tabelaDeParticipantes);
//		oos.flush();
//		ByteBuffer bf = ByteBuffer.allocate(baos.size()+1);
//		bf.put((byte) 5);
//		bf.put(baos.toByteArray());
//		bf.flip();
//		AppSocket sock = this.connectToClient(ci, listener);
//		if(sock == null){
//			sock = tabelaDeSockets.get(ci.getIp());
//		}
//		ChatThread thread = listaDeThreads.get(ci.getIp());
//		thread.addGroupListener(groupListener);
//		sock.send(bf.array());

	}
	public void sendGroup(String msg, Hashtable<ConnectionInfo, String> tabelaDeParticipantes, ClientGuiListener listener) throws GroupException{
		byte[] rawMsg = null;
		String logins = "";
		try {
			rawMsg = msg.getBytes("UTF-16");
		} catch (UnsupportedEncodingException e) {
			rawMsg = msg.getBytes();
			System.out.println("Charset no suportado");
		}
		ByteBuffer pacote = ByteBuffer.allocate(rawMsg.length+5);
		pacote.put((byte)4);
		pacote.put(ByteOp.intToByte(rawMsg.length));
		pacote.put(rawMsg);
		pacote.flip();
		
		Enumeration<ConnectionInfo> peers = tabelaDeParticipantes.keys();
		
		while(peers.hasMoreElements()){
			ConnectionInfo temp = peers.nextElement();
			AppSocket sock = tabelaDeSockets.get(temp.getIp());
			try{
				if (sock != null) {
					sock.send(pacote.array());
				}else{
					sock = this.connectToClient(temp, listener);
					sock.send(pacote.array());
				}
			} catch (TimeoutLimitException e) {
				logins += " " + tabelaDeParticipantes.get(temp);
			}
		}
		if(!logins.equals("")){
			throw new GroupException("No foi possvel enviar a mensagem para os seguintes logins:\n"+logins);
		}
	}
	private synchronized AppSocket connectToClient(ConnectionInfo ci, ClientGuiListener listener) throws TimeoutLimitException{
		AppSocket retorno = null;
		if(!tabelaDeSockets.containsKey(ci.getIp())){
			ci.setId(0);
			AppSocket sock = new AppSocket(trans, nextId);
			sock.connect(ci);
			ChatThread thread = new ChatThread(this, sock, ci, listener);
			listaDeThreads.put(ci.getIp(), thread);
			System.out.println(">>>>>>>>>>>>>>>>>>>>>LISTENER");
			thread.addListener(listener);
			nextId++;
			tabelaDeSockets.put(ci.getIp(), sock);
			retorno = sock;
		}
		return retorno;
	}
	private synchronized AppSocket connectToClient(ConnectionInfo ci, ClientGuiListener listener, ClientGuiListener groupListener) throws TimeoutLimitException{
		AppSocket retorno = null;
//		if(!tabelaDeSockets.containsKey(ci.getIp())){
//			ci.setId(0);
//			AppSocket sock = new AppSocket(trans, nextId);
//			sock.connect(ci);
//			ChatThread thread = new ChatThread(this, sock, ci);
//			listaDeThreads.put(ci.getIp(), thread);
//			thread.addListener(listener);
//			nextId++;
//			tabelaDeSockets.put(ci.getIp(), sock);
//			retorno = sock;
//		}
		return retorno;
	}
	
	public void connectToGroup(Enumeration<ConnectionInfo> ips,ClientGuiListener listener, ClientGuiListener groupListener) throws TimeoutLimitException{
//		while(ips.hasMoreElements()){
//			ConnectionInfo peer = ips.nextElement();
//			if(!tabelaDeSockets.contains(peer.getIp())){
//				peer.setId(0);
//				AppSocket sock = new AppSocket(trans, nextId);
//				nextId++;
//				sock.connect(peer);
//				ChatThread thread = new ChatThread(this, sock, peer);
//				listaDeThreads.put(peer.getIp(), thread);
//				thread.addListener(listener);
//				thread.addGroupListener(groupListener);
//				tabelaDeSockets.put(peer.getIp(), sock);
//			
//			}else{
//				ChatThread temp = listaDeThreads.get(peer.getIp());
//				if(temp != null){
//					temp.addGroupListener(groupListener);
//				}
//			}
//		}
	}
}
